# Copyright (c) 2010-2011,2013 Turbulenz Limited
"""
Python module that handles JSON asset disassembling into HTML.
"""

#######################################################################################################################

def ordered(d):
    keys = d.keys()
    keys.sort()
    for k in keys:
        yield(k, d[k])

# pylint: disable=R0201
class Json2txtRenderer(object):

    def __init__(self):
        pass

    def span(self, string, span_class=None):
        return string

    def key(self, string):
        return string

    def comment(self, string):
        return string

    ###################################

    def expand(self, node_path_string, expand_all=True):
        return '...'

    def collapse(self):
        return ''

    def node_span(self, node_path_string):
        return ''

    def close_span(self):
        return ''

    ###################################

    def expand_link(self, num, is_list, node_path_string, parent):
        """Creates a link span to expand a dictionary or a list."""
        return '%s%s%s%s%s\n%s' % ( self.node_span(node_path_string),
                                    '[ ' if is_list else '{ ',
                                    self.expand(node_path_string, False),
                                    self.comment(' (%i %s)' % (num, parent)),
                                    ' ]' if is_list else ' }',
                                    self.close_span() )

    def string(self, value, link_prefix=None, link=False):
        asset = unicode(value)
        return asset


class Json2txtColourRenderer(Json2txtRenderer):

    def __init__(self):
        Json2txtRenderer.__init__(self)

    def span(self, string, span_class=None):
        return string

    def key(self, string):
        return '\033[31m%s\033[0m' % string

    def comment(self, string):
        return '\033[34m%s\33[0m' % string

    def string(self, value, link_prefix=None, link=False):
        return '\033[32m"%s"\033[0m' % unicode(value)

    def expand(self, node_path_string, expand_all=True):
        return '\033[34m...\033[0m'


class Json2htmlRenderer(Json2txtRenderer):

    def __init__(self):
        Json2txtRenderer.__init__(self)

    def span(self, string, span_class=None):
        if span_class:
            return '<span class="%s">%s</span>' % (span_class, string)
        else:
            return '<span>%s</span>' % string

    def key(self, string):
        return self.span(string)

    def comment(self, string):
        return self.span(string, 'c')

    ###################################

    def expand(self, node_path_string, expand_all=True):
        c = 'expand c all' if expand_all else 'expand c'
        return '<a class="%s">more</a>' % c

    def collapse(self):
        return '<a class="collapse c">less</a>'

    def node_span(self, node_path_string):
        return '<span class="node" id="node=%s">' % node_path_string

    def close_span(self):
        return '</span>'

    ###################################

    def string(self, value, link_prefix=None, link=False):
        asset = unicode(value)
        if link_prefix and (link or '/' in value):
            return '"<a href="%s/%s">%s</a>"' % (link_prefix, asset, asset)
        else:
            return '"%s"' % asset
# pylint: enable=R0201

#######################################################################################################################

class Disassembler:
    """Convert JSON to HTML."""

    def __init__(self, renderer, list_cull=3, dict_cull=3, depth=2, link_prefix='',
                 single_line_string_length=200, auto_expand_child_lenght=500, limit_list_length=1000):
        self.renderer = renderer
        self.list_cull = list_cull
        self.dict_cull = dict_cull
        self.depth = depth

        self.single_line_string_length = single_line_string_length
        self.auto_expand_child_lenght = auto_expand_child_lenght
        self.limit_list_length = limit_list_length

        self.link_prefix = link_prefix

        self.current_node_path = [ ]
        self.node_path_string = ''
        self.current_depth = 0

    def _update_node_path_string(self):
        self.node_path_string = ','.join([str(x) for x in self.current_node_path])

    def _push(self, index):
        self.current_node_path.append(index)
        self._update_node_path_string()
        self.current_depth += 1

    def _pop(self):
        self.current_node_path = self.current_node_path[:-1]
        self._update_node_path_string()
        self.current_depth -= 1

    def _has_more_depth(self):
        return self.current_depth <= self.depth

    def _indents(self):
        return ('  ' * len(self.current_node_path), '  ' * (len(self.current_node_path) - 1))

    ###################################################################################################################

    # This function can render the list on multiple lines.
    def mark_up_list_items(self, output, element, start_element, count):
        """Iterate through the list or its slice and mark up its elements."""
        (indent, minor_indent) = self._indents()
        r = self.renderer

        def _expand_link(length, is_list):
            return r.expand_link(length, is_list, self.node_path_string, 'items' if is_list else 'elements')

        sub_list = element[start_element:start_element + count]
        for i, l in enumerate(sub_list):
            if i < (len(sub_list) - 1):
                comma = ','
                line_indent = indent
            else:
                comma = ''
                line_indent = minor_indent

            self._push(i + start_element)
            if isinstance(l, dict):
                if self._has_more_depth():
                    self.mark_up_dict(output, l)
                else:
                    child_output = [ ]
                    self.mark_up_dict(child_output, l)
                    child_len = reduce(lambda a, b: a + len(b), child_output, 0)
                    if child_len < self.auto_expand_child_lenght:
                        output.extend(child_output)
                    else:
                        output.append(_expand_link(len(l), False))
            elif isinstance(l, (list, tuple)):
                if self._has_more_depth():
                    self.mark_up_list(output, l)
                else:
                    child_output = [ ]
                    self.mark_up_list(child_output, l)
                    child_len = reduce(lambda a, b: a + len(b), child_output, 0)
                    if child_len < self.auto_expand_child_lenght:
                        output.extend(child_output)
                    else:
                        output.append(_expand_link(len(l), True))
            elif isinstance(l, (str, unicode)):
                output.append('%s%s\n' % (r.string(l, self.link_prefix), comma))
            elif isinstance(l, bool):
                output.append('%s%s\n' % ('true' if l else 'false', comma))
            elif l is None:
                output.append('null%s\n' % comma)
            else:
                output.append('%s%s\n' % (str(l), comma))
            output.append(line_indent)
            self._pop()

    def mark_up_single_line_list_items(self, output, element, start_element, count):
        # This function can render on a single line the list.
        (indent, _) = self._indents()
        r = self.renderer
        limit_list_length = self.limit_list_length

        current_length = 0
        for i, x in enumerate(element[start_element:start_element + count]):
            if isinstance(x, (str, unicode)):
                o = r.string(x, self.link_prefix)
            elif isinstance(x, bool):
                o = 'true' if x else 'false'
            elif x is None:
                o = 'null'
            else:
                o = str(x)

            output.append(o)
            if i < (count - 1):
                output.append(', ')

            # We count the size of the element and an extra 2 for the comma and space.
            current_length += len(o) + 2
            if current_length > limit_list_length:
                output.append('\n%s' % indent)
                current_length = 0

    def mark_up_list(self, output, element, parent=None, expand=False):
        (indent, _) = self._indents()
        r = self.renderer

        num_values = len(element)
        over_size_limit = num_values > 2 * self.list_cull

        parent = parent or 'items'

        # Test the list to see if we can render it onto a single line.
        # If we find a dict or list we exit and fall back onto rendering each item on a single line.
        total_length = 0
        for x in element:
            if isinstance(x, (dict, list, tuple)):
                break
            elif isinstance(x, (str, unicode)):
                total_length += len(x)
        else:
            # We got to the end of the list without finding any complicated items.
            # If the total length of strings too large also fall back onto multi line rendering.
            if total_length < self.single_line_string_length:
                if not expand and over_size_limit:
                    output.append('%s[' % r.node_span(self.node_path_string))
                    self.mark_up_single_line_list_items(output, element, 0, self.list_cull)
                    output.append(', %s ' % r.expand(self.node_path_string))
                    self.mark_up_single_line_list_items(output, element, num_values - self.list_cull, self.list_cull)
                    output.append(']%s\n%s' % (r.comment(' (%i of %i %s)' % (self.list_cull * 2, num_values, parent)),
                                               r.close_span()))
                else:
                    output.append('%s[' % ((r.collapse() + ' ') if over_size_limit else ''))
                    self.mark_up_single_line_list_items(output, element, 0, num_values)
                    output.append(']\n%s' % (r.close_span() if over_size_limit else ''))

                # Early out as we've followed a special case.
                return

        # If we get this far we render each item on it's own line.
        if not expand and over_size_limit:
            # display start and end of the list
            output.append('%s[\n%s' % (r.node_span(self.node_path_string), indent))
            self.mark_up_list_items(output, element, 0, self.list_cull)
            output.append('%s\n%s' % (r.expand(self.node_path_string), indent))
            self.mark_up_list_items(output, element, num_values - self.list_cull, self.list_cull)
            output.append(']%s\n%s' % (r.comment(' (%i of %i %s)' % (self.list_cull * 2, num_values, parent)),
                                       r.close_span()))
        else:
            output.append('%s[\n%s' % ((r.collapse() + ' ') if over_size_limit else '', indent))
            self.mark_up_list_items(output, element, 0, num_values)
            output.append(']\n')

    def mark_up_element(self, output, k, v, i=0):
        """Mark up the element of a node."""
        (indent, _) = self._indents()
        r = self.renderer

        def _expand_link(is_list):
            return r.expand_link(len(v), is_list, self.node_path_string, k)

        self._push(i)
        output.append('%s%s: ' % (indent, r.key(k)) if k is not None else '')
        if isinstance(v, dict):
            if self._has_more_depth():
                self.mark_up_dict(output, v, k)
            else:
                child_output = [ ]
                self.mark_up_dict(child_output, v, k)
                child_len = reduce(lambda a, b: a + len(b), child_output, 0)
                if child_len < self.auto_expand_child_lenght:
                    output.extend(child_output)
                else:
                    output.append(_expand_link(False))
        elif isinstance(v, (list, tuple)):
            if self._has_more_depth():
                self.mark_up_list(output, v, k)
            else:
                child_output = [ ]
                self.mark_up_list(child_output, v, k)
                child_len = reduce(lambda a, b: a + len(b), child_output, 0)
                if child_len < self.auto_expand_child_lenght:
                    output.extend(child_output)
                else:
                    output.append(_expand_link(True))
        elif isinstance(v, (str, unicode)):
            output.append('%s\n' % r.string(v, self.link_prefix, (k == 'reference')))
        elif isinstance(v, (bool)):
            output.append('true\n' if v is True else 'false\n')
        elif v is None:
            output.append('none\n')
        else:
            output.append('%s\n' % str(v))
        self._pop()

    def mark_up_dict(self, output, element, parent=None, expand=False):
        """Mark up the node element accordingly to its type into HTML and return as a string."""
        (indent, minor_indent) = self._indents()
        r = self.renderer

        num_values = len(element)
        over_size_limit = num_values > self.dict_cull

        parent = parent or 'elements'

        if not expand and over_size_limit:
            output.append('%s{\n' % r.node_span(self.node_path_string))

            for i, (k, v) in enumerate(ordered(element)):
                if i == self.dict_cull:
                    break
                self.mark_up_element(output, k, v, i)

            output.append('%s%s %s\n%s}\n%s' % (indent,
                                               r.expand(self.node_path_string),
                                               r.comment('(%i of %i %s)' % (self.dict_cull, num_values, parent)),
                                               minor_indent,
                                               r.close_span()))

        else:
            output.append('%s{\n' % ((r.collapse() + ' ') if over_size_limit else ''))
            for i, (k, v) in enumerate(ordered(element)):
                self.mark_up_element(output, k, v, i)
            output.append('%s}\n' % minor_indent)

    def find_node(self, output, json_asset, node_list, expand=False):
        """Find the element in the node list from the json_asset, return marked up in HTML."""

        def _values(elements):
            if isinstance(json_asset, dict):
                for k, v in ordered(json_asset):
                    yield (k, v)
            elif isinstance(json_asset, list):
                for v in json_asset:
                    yield (None, v)

        for i, (k, v) in enumerate(_values(json_asset)):
            if i == node_list[0]:
                self._push(i)
                if (len(node_list) == 1):
                    if isinstance(v, dict):
                        self.mark_up_dict(output, v, k, expand)
                    elif isinstance(v, list):
                        self.mark_up_list(output, v, k, expand)
                    else:
                        # Potentially this should handle the other types correct.
                        # Maybe we could refactor this out of the dict or list methods.
                        output.append('%s\n' % str(v))
                else:
                    self.find_node(output, v, node_list[1:], expand)
                self._pop()

    def mark_up_asset(self, json_asset, expand=False, node=None):
        """Mark up element on HTTP Request."""
        if node is None:
            node_list = [0]
        else:
            node_list = [int(x) for x in node.split(',')]

        output = [ ]
        self.find_node(output, json_asset, node_list, expand)
        return ''.join(output)



#!/usr/bin/env python
# Copyright (c) 2013 Turbulenz Limited

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from turbulenz_tools import __version__

import platform
import sys
from glob import iglob

if 'sdist' in sys.argv:
    SCRIPTS = list(iglob('scripts/*'))
elif platform.system() == 'Windows':
    SCRIPTS = list(iglob('scripts/*.bat'))
else:
    SCRIPTS = [ s for s in iglob('scripts/*') if not s.endswith('.bat') ]

setup(name='turbulenz_tools',
    version=__version__,
    description='Tools for the creation of games with the Turbulenz Engine',
    author='Turbulenz Limited',
    author_email='support@turbulenz.com',
    url='https://turbulenz.com/',
    install_requires=[
        'simplejson>=2.1.5',
        'jinja2>=2.4',
        'PyYAML>=3.09',
        'jsmin>=2.0.2'
        ],
    scripts=SCRIPTS,
    packages=[ 'turbulenz_tools', 'turbulenz_tools.tools', 'turbulenz_tools.utils' ],
    include_package_data=True,
    package_data={},
    zip_safe=False,
    license = 'MIT',
    platforms = 'Posix; MacOS X; Windows',
    classifiers = ['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'License :: OSI Approved :: MIT License',
                   'Operating System :: OS Independent',
                   'Topic :: Software Development',
                   'Programming Language :: Python :: 2.7'],
    )



#!/usr/bin/python
# Copyright (c) 2009-2013 Turbulenz Limited
"""
Convert LightWave (.obj) OBJ2 files into a Turbulenz JSON asset.

Supports generating NBTs.
"""

import logging
LOG = logging.getLogger('asset')

# pylint: disable=W0403
from stdtool import standard_main, standard_json_out, standard_include
from asset2json import JsonAsset
from mesh import Mesh
from node import NodeName
from os.path import basename
# pylint: enable=W0403

__version__ = '1.2.0'
__dependencies__ = ['asset2json', 'mesh', 'node', 'vmath']


DEFAULT_EFFECT_NAME = 'lambert'

# Note:
# * Does not have support for .mtl files yet, but expects any relevant materials
#   to be declared in a .material file and included as a dependancy in deps.yaml
# * This script sets the default node name of the asset parsed to be the file name (without the path),
#   unless anything else is supplied. This could lead to clashes with other nodes with the same name.
# * Each surface is assumed to only have a single material. A new surface will be made upon requiring a new material.

#######################################################################################################################

def _increment_name(name):
    """Returns a name similar to the inputted name

        If the inputted name ends with a hyphen and then a number,
        then the outputted name has that number incremented.
        Otherwise, the outputted name has a hyphen and the number '1' appended to it.
    """
    index = name.rfind('-')
    if index == -1:
        # The is not already followed by a number, so just append -1
        return name + '-1'
    else:
        # The number is already followed by a number, so increment the number
        value = int(name[index+1:])
        return name[:index+1] + str(value+1)

def purge_empty(dictionary, recurseOnce = False):
    """Removes all elements of a dictionary which return true for elem.is_empty().

        If recurseOnce is True, then it will call element.purge_empty()
        for any remaining elements in the dictionary. This will not recurse further."""
    names_empty = []
    for name, elem in dictionary.iteritems():
        if elem.is_empty():
            names_empty.append(name)
    # Need to compile the list of names of empty elements in a separate loop above,
    # to avoid changing dictionary during iteration over it directly.
    for name in names_empty:
        del dictionary[name]
    if recurseOnce:
        for elem in dictionary.itervalues():
            # Note that it is here assumed that the element has a method purge_empty
            elem.purge_empty()


#######################################################################################################################

class Surface:
    """Represents a surface (a.k.a. group) as parsed from the .obj file into .json format

        Contains a material, and indices into the Obj2json.primitives list
        to identify which polygons belong to this surface."""
    def __init__(self, first, material):
        # first refers to initial index of triangles parsed
        # count refers to number of triangles parsed
        self.first = first
        self.count = 0
        self.material_name = material
    def is_empty(self):
        return self.count == 0

class Shape:
    """Represents a shape (a.k.a. object) as parsed from the .obj file into .json format

        Contains a dictionary of names -> surfaces."""
    # NB: Do not pass in Null for the surfaces parameter, all shapes are assumed to
    #     contain a dictionary with at least one surface, even if it is a default one.
    def __init__(self, surfaces):
        # Dictionary of names -> surfaces
        self.surfaces = surfaces
    def is_empty(self):
        empty = True
        for surface in self.surfaces.itervalues():
            # If a single surface is non-empty, empty will be false
            empty = empty and surface.is_empty()
        return empty
    def purge_empty(self):
        purge_empty(self.surfaces)

#######################################################################################################################

# pylint: disable=R0904
class Obj2json(Mesh):
    """Parse a OBJ file and generate a Turbulenz JSON geometry asset."""
    # TODO: Probably some more information should be moved from the Obj2json class to the Shape class.

    def __init__(self, obj_file_name):
        self.default_shape_name = obj_file_name or 'initial_shape'
        self.default_surface_name = 'initial_surface'
        self.default_material_name = 'default'

        # Name of the shape/surface/material from the most recently parsed 'o'/'g'/'usemtl' line respectively
        self.curr_shape_name = self.default_shape_name
        self.curr_surf_name = self.default_surface_name
        self.curr_mtl_name = self.default_material_name

        # To keep of track of number of polygons parsed so far
        self.next_polygon_index = 0
        # A tuple of indices per vertex element
        self.indices = [ ]

        # Shortcut to be able to access the current surface in fewer characters and lookups
        self.curr_surf = Surface(0, self.default_material_name)
        # Dictionary of names -> shapes. Initialise to a default shape with a default surface
        self.shapes = { self.default_shape_name : Shape({ self.default_surface_name : self.curr_surf }) }
        Mesh.__init__(self)

    def __read_object_name(self, data):
        """Parse the 'o' line. This line contains the mesh name."""
        LOG.debug("object name:%s" % data)
        # Remove any shapes with no surfaces (e.g. the default shape if a named one is given)
        purge_empty(self.shapes)
        # If a shape with this name has already been declared, do not change it,
        # append the following faces to the current shape in stead
        if data not in self.shapes:
            self.curr_shape_name = data
            self.curr_surf = Surface(self.next_polygon_index, self.curr_mtl_name)
            self.shapes[data] = Shape({self.curr_surf_name : self.curr_surf})

    def __read_group_name(self, data):
        """Parse the 'g' line. This indicates the start of a group (surface)."""
        # Remove leading/trailing whitespace
        data = data.strip()
        LOG.debug("group name:%s" % data)
        # Note: Don't purge empty shapes/surfaces here, you might remove a new surface
        #       created by a preceding 'usemtl' line. Purging happens after parsing.
        self.curr_surf_name = data
        # Use most recently specified material (unless overridden later)
        self.curr_surf = Surface(self.next_polygon_index, self.curr_mtl_name)
        self.shapes[self.curr_shape_name].surfaces[data] = self.curr_surf

    def __read_material(self, data):
        """Parse the 'usemtl' line. This references a material."""
        data = data.strip()
        LOG.debug("material name:%s" % data)
        self.curr_mtl_name = data
        if self.curr_surf.is_empty():
            # No polygons (yet) in the current surface, so just set its material
            self.curr_surf.material_name = data
        elif self.curr_surf.material_name != data:
            # Current surface already has a number of faces of a different material
            # so create a new surface for this new material
            self.curr_surf = Surface(self.next_polygon_index, data)
            self.curr_surf_name = _increment_name(self.curr_surf_name)
            self.shapes[self.curr_shape_name].surfaces[self.curr_surf_name] = self.curr_surf

    def __read_vertex_position(self, data):
        """Parse the 'v' line. This line contains the vertex position."""
        sv = data.split(' ')
        position = (float(sv[0]), float(sv[1]), float(sv[2]))
        self.positions.append(position)
        # Do not calculate the bounding box here, as some unused vertices may later be removed

    def __read_vertex_uvs(self, data):
        """Parse the 'vt' line. This line contains the vertex uvs."""
        # Texture coordinates
        sv = data.split(' ')
        if len(sv) == 2:
            uvs = (float(sv[0]), float(sv[1]))
        else:
            uvs = (float(sv[0]), float(sv[1]), float(sv[2]))
        self.uvs[0].append(uvs)

    def __read_vertex_normal(self, data):
        """Parse the 'vn' line. This line contains the vertex normals."""
        (sv0, sv1, sv2) = data.split(' ')
        normal = (float(sv0), float(sv1), -float(sv2))
        self.normals.append(normal)

    def __read_face(self, data):
        """Parse the 'f' line. This line contains a face.

            Constructs a tri-fan if face has more than 3 edges."""
        def __extract_indices(si):
            """Add a tuple of indices."""
            # Vertex index / Texture index / Normal index
            # Subtract 1 to count indices from 0, not from 1.
            s = si.split('/')
            if len(s) == 1:
                return [int(s[0]) - 1]
            if len(s) == 2:
                return (int(s[0]) - 1, int(s[1]) - 1)
            else:
                if len(s[1]) == 0:
                    return (int(s[0]) - 1, int(s[2]) - 1)
                else:
                    return (int(s[0]) - 1, int(s[1]) - 1, int(s[2]) - 1)

        # Split string into list of vertices
        si = data.split()
        indices = self.indices
        # Construct a tri-fan of all the vertices supplied (no support for quadrilaterals or polygons)
        # Origin vertex of fan
        i0 = __extract_indices(si[0])
        prevInd = __extract_indices(si[1])
        for i in xrange(2, len(si)):
            currInd = __extract_indices(si[i])
            indices.append(i0)
            indices.append(prevInd)
            indices.append(currInd)
            prevInd = currInd
        num_triangles = len(si) - 2
        self.next_polygon_index += num_triangles
        self.curr_surf.count += num_triangles

    def __ignore_comments(self, data):
        """Ignore comments."""

#######################################################################################################################

    def parse(self, f, prefix = ""):
        """Parse an OBJ file stream."""
        chunks_with_data = { 'v': Obj2json.__read_vertex_position,
                             'vt': Obj2json.__read_vertex_uvs,
                             'vn': Obj2json.__read_vertex_normal,
                             'f': Obj2json.__read_face,
                             'o': Obj2json.__read_object_name,
                             'g': Obj2json.__read_group_name,
                             'usemtl': Obj2json.__read_material,
                             '#': Obj2json.__ignore_comments}
        for lineNumber, line in enumerate(f):
            # The middle of the tuple is just whitespace
            (command, _, data) = line.partition(' ')
            if len(data) > 0:
                data = data[:-1]
                while len(data) > 0 and data[0] == ' ':
                    data = data[1:]
                if len(data) > 0:
                    # After stripping away excess whitespace
                    address_string = "(%d) %s%s" % (lineNumber, prefix, command)
                    if command in chunks_with_data:
                        LOG.debug(address_string)
                        # Parse data depending on its type
                        chunks_with_data[command](self, data)
                    else:
                        LOG.warning(address_string + " *unsupported*")

    def unpack_vertices(self):
        """Unpack the vertices."""
        # Consecutive list of nodes making up faces (specifically, triangles)
        indices = []
        # A node of a face definition consists of a vertex index, and optional
        # texture coord index and an optional normal vector index
        # Thus, the length of an element of self.indices can be 1, 2 or 3.
        if len(self.indices[0]) == 1:
            # No texture coordinates (uv) or normal vector specified.
            indices = [x[0] for x in self.indices]
        else:
            old_positions = self.positions
            old_uvs = self.uvs[0]
            old_normals = self.normals
            positions = []  # Vertex position
            uvs = []        # Texture coordinate
            normals = []
            mapping = {}
            if len(self.indices[0]) == 2:
                for indx in self.indices:
                    hash_string = "%x:%x" % (indx[0], indx[1])
                    if hash_string in mapping:
                        indices.append(mapping[hash_string])
                    else:
                        newindx = len(positions)
                        mapping[hash_string] = newindx
                        indices.append(newindx)
                        positions.append(old_positions[indx[0]])
                        # Figure out whether 2nd value is uv or normal
                        if len(old_uvs) != 0:
                            uvs.append(old_uvs[indx[1]])
                        else:
                            normals.append(old_normals[indx[1]])
            else:
                for indx in self.indices:
                    hash_string = "%x:%x:%x" % (indx[0], indx[1], indx[2])
                    if hash_string in mapping:
                        indices.append(mapping[hash_string])
                    else:
                        newindx = len(positions)
                        mapping[hash_string] = newindx
                        indices.append(newindx)
                        positions.append(old_positions[indx[0]])
                        uvs.append(old_uvs[indx[1]])
                        normals.append(old_normals[indx[2]])
            # Reassign the vertex positions, texture coordinates and normals, so
            # that they coincide with the indices defining the triangles.
            self.positions = positions
            self.uvs[0] = uvs
            self.normals = normals
        self.generate_primitives(indices)

    def extract_nbt_options(self, definitions_asset):
        """Returns whether normals and tangents/binormals are needed, and whether they should be generated.

            Loops over each material and checks their meta attributes to extract this information."""
        # Record the whether normals/tangents need to be generated, and which shapes require these options
        generate_normals  = False
        generate_tangents = False
        need_normals      = set()
        need_tangents     = set()
        for shape_name in self.shapes.iterkeys():
            for surface_name in self.shapes[shape_name].surfaces.iterkeys():
                material_name = self.shapes[shape_name].surfaces[surface_name].material_name
                material = definitions_asset.retrieve_material(material_name, default = True)
                effect = definitions_asset.retrieve_effect(material['effect'])

                # Rules used: Generating tangents implies needing tangents
                #             Needing tangents implies needing normals
                #             Needing tangents/normals implies generating them if they aren't present
                if material.meta('generate_tangents') or effect is not None and effect.meta('generate_tangents'):
                    generate_tangents = True
                    need_tangents.add(shape_name)
                elif material.meta('tangents') or effect is not None and effect.meta('tangents'):
                    need_tangents.add(shape_name)
                    # Generate tangents if any material needs tangents and you haven't parsed any,
                    # or if any materials ask you to generate tangents
                    generate_tangents = generate_tangents or not len(self.tangents) or not len(self.binormals)
                if material.meta('generate_normals') or effect is not None and effect.meta('generate_normals'):
                    generate_normals = True
                    need_normals.add(shape_name)
                elif material.meta('normals') or effect is not None and effect.meta('normals'):
                    need_normals.add(shape_name)
                    # Same reasoning as with generating tangents
                    generate_normals = generate_normals or not len(self.normals)
        return (need_normals, generate_normals, need_tangents, generate_tangents)
# pylint: enable=R0904

#######################################################################################################################

def parse(input_filename="default.obj", output_filename="default.json", asset_url="", asset_root=".",
          infiles=None, options=None):
    """Utility function to convert an OBJ file into a JSON file."""
    definitions_asset = standard_include(infiles)
    with open(input_filename, 'rt') as source:
        asset = Obj2json(basename(input_filename))
        asset.parse(source)
        # Remove any and all unused (e.g. default) shapes and surfaces
        purge_empty(asset.shapes, recurseOnce = True)
        # Generate primitives
        asset.unpack_vertices()
        # Remove any degenerate primitives unless they're requested to be kept
        keep_degenerates = False
        for shape in asset.shapes:
            for _, surface in asset.shapes[shape].surfaces.iteritems():
                material = definitions_asset.retrieve_material(surface.material_name)
                if material.meta('keep_degenerates'):
                    keep_degenerates = True
        if not keep_degenerates:
            asset.remove_degenerate_primitives()
        # Remove any unused vertices and calculate a bounding box
        asset.remove_redundant_vertexes()
        asset.generate_bbox()
        # Generate normals/tangents if required
        (need_normals, generate_normals,
         need_tangents, generate_tangents) = asset.extract_nbt_options(definitions_asset)
        if generate_tangents:
            if generate_normals:
                asset.generate_normals()
            asset.generate_smooth_nbts()
            asset.invert_v_texture_map()
        elif generate_normals:
            asset.generate_normals()
            asset.smooth_normals()
        json_asset = JsonAsset()
        for shape_name in asset.shapes.iterkeys():
            json_asset.attach_shape(shape_name)
            node_name = NodeName("node-%s" % shape_name)
            json_asset.attach_node(node_name)
            #TODO: Should the following be divided into separate shapes?
            json_asset.attach_positions(asset.positions, shape_name)
            # Attach texture map, normals and tangents/binormals if required
            if len(asset.uvs[0]) != 0:
                json_asset.attach_uvs(asset.uvs[0], shape_name)
            if shape_name in need_tangents:
                if len(asset.tangents):
                    # Needing tangents implies needing normals and binormals
                    json_asset.attach_nbts(asset.normals, asset.tangents, asset.binormals, shape_name)
                else:
                    LOG.error('tangents requested for shape %s, but no tangents or uvs available!' % shape_name)
            elif shape_name in need_normals:
                json_asset.attach_normals(asset.normals, shape_name)
            for surface_name, surface in asset.shapes[shape_name].surfaces.iteritems():
                material = definitions_asset.retrieve_material(surface.material_name)
                effect = material.get('effect', DEFAULT_EFFECT_NAME)
                effect_name = "effect-%s" % shape_name
                material_name = "material-%s" % surface.material_name
                instance_name = "instance-%s-%s" % (shape_name, surface_name)
                json_asset.attach_effect(effect_name, effect)
                mat_params = material.get('parameters', None)
                json_asset.attach_material(material_name, effect=effect, parameters=mat_params)
                def textures(mat_params):
                    for k, v in mat_params.iteritems():
                        # If a paramater of a material has a string value, it is assumed to be a texture definition
                        if isinstance(v, basestring):
                            # Return the type of the texture (e.g. 'diffuse')
                            yield k
                for t_type in textures(mat_params):
                    json_asset.attach_texture(material_name, t_type, mat_params[t_type])
                first = surface.first
                last = first + surface.count
                json_asset.attach_surface(asset.primitives[first:last], JsonAsset.SurfaceTriangles,
                                          shape_name, name=surface_name)
                json_asset.attach_node_shape_instance(node_name, instance_name, shape_name,
                                                      material_name, surface=surface_name)
        json_asset.attach_bbox(asset.bbox)
        standard_json_out(json_asset, output_filename, options)
        return json_asset

if __name__ == "__main__":
    standard_main(parse, __version__,
                  "Convert LightWave (.obj) OBJ2 files into a Turbulenz JSON asset. Supports generating NBTs.",
                  __dependencies__)



# Copyright (c) 2010-2013 Turbulenz Limited
"""
Utility functions for finding and outputing dependencies
"""

from os.path import join, exists, abspath
from jinja2 import meta

__version__ = '1.0.0'

def find_file_in_dirs(filename, dirs, error_on_multiple = False):
    found = []
    for d in dirs:
        fn = join(d, filename)
        if exists(fn):
            if error_on_multiple:
                if fn not in found:
                    found.append(fn)
            else:
                return abspath(fn)
    if error_on_multiple:
        if len(found) > 1:
            raise Exception("File '%s' matched several locations: %s" % found)
        if len(found) == 1:
            return abspath(found[0])
    return None

# pylint: disable=W0102
def find_dependencies(input_file, templatedirs, env, exceptions=[]):
    """
    Find jinja2 dependency list.  Files listed in 'exceptions' do not
    generate exceptions if not found.
    """

    total_set = set()

    def find_dependencies_recurse(file_path):
        new_deps = []

        # Parse the file and extract the list of references

        with open(file_path) as f:
            ast = env.parse(f.read())

            # For each reference, find the absolute path.  If no file
            # is found and the reference was not listed in exceptions,
            # throw an error.

            for reference in meta.find_referenced_templates(ast):
                reference_path = find_file_in_dirs(reference, templatedirs)
                if reference_path is None:
                    if reference in exceptions:
                        continue
                    raise Exception("cannot find file '%s' referenced in "
                                    "'%s'" % (reference, file_path))
                new_deps.append(reference_path)

        for dep in new_deps:
            # Make sure we don't have a circular reference
            if dep not in total_set:
                total_set.add(dep)
                find_dependencies_recurse(dep)
        return

    top_file = find_file_in_dirs(input_file, templatedirs)
    if top_file is None:
        raise Exception("cannot find file '%s'" % input_file)
    total_set.add(top_file)
    find_dependencies_recurse(top_file)

    sorted_total = []
    for x in total_set:
        sorted_total.append(x)

    sorted_total.sort()
    return sorted_total
# pylint: enable=W0102



# Copyright (c) 2012-2013 Turbulenz Limited
"""
Keep track of the cost of various points in code
"""

import time

############################################################

class ResultNode(object):
    def __init__(self, node_name):
        self.name = node_name
        self.duration = -1
        self.children = []

        self._start = time.time()

    def stop(self):
        self.duration = time.time() - self._start

    def add_child(self, child_result):
        if -1 != self.duration:
            raise Exception("section '%s' already stopped when trying to add "
                            "child '%s'" % (self.name, child_result.name))
        self.children.append(child_result)

############################################################

class ProfilerDummyImpl(object):
    @classmethod
    def __init__(cls):
        return
    @classmethod
    def start(cls, _):
        return
    @classmethod
    def stop(cls, _):
        return
    @classmethod
    def get_root_nodes(cls):
        return []
    @classmethod
    def dump_data(cls):
        return

############################################################

class ProfilerImpl(object):

    def __init__(self):
        self._root = ResultNode('__')
        self._current_node = self._root
        self._current_stack = []

    def start(self, section_name):
        new_child = ResultNode(section_name)
        self._current_node.add_child(new_child)

        self._current_stack.append(self._current_node)
        self._current_node = new_child

    def stop(self, section_name):

        # Unwind stack until we find the section
        while section_name != self._current_node.name:
            if len(self._current_stack) == 0:
                raise Exception("Cannot find section '%s' to stop it" \
                                    % section_name)

            self._current_node = self._current_stack.pop()

        self._current_node.stop()
        self._current_node = self._current_stack.pop()

    def get_root_nodes(self):
        return self._root.children

    def dump_data(self):

        if 0 == len(self._root.children):
            return

        def _dump_node(node, _indent = 2):
            if node.duration == -1:
                duration = "(unterminated)"
            else:
                duration = node.duration
            _indent_string = " "*_indent
            print "%s%-16s - %s%.6f" % (_indent_string, node.name, _indent_string,
                                    duration)
            for c in node.children:
                _dump_node(c, _indent+2)

        print "TimingData: "
        for r in self._root.children:
            _dump_node(r)

############################################################

class Profiler(object):

    _profiler_impl = ProfilerDummyImpl()

    @classmethod
    def enable(cls):
        if not isinstance(cls._profiler_impl, ProfilerDummyImpl):
            raise Exception("Profiler.enable_profiler() called twice")
        cls._profiler_impl = ProfilerImpl()

    @classmethod
    def start(cls, section_name):
        cls._profiler_impl.start(section_name)

    @classmethod
    def stop(cls, section_name):
        cls._profiler_impl.stop(section_name)

    @classmethod
    def get_root_nodes(cls):
        return cls._profiler_impl.get_root_nodes()

    @classmethod
    def dump_data(cls):
        cls._profiler_impl.dump_data()

############################################################

def _profiler_test():

    ##################################################

    p = ProfilerImpl()
    p.start('section1')
    p.stop('section1')
    roots = p.get_root_nodes()
    assert(1 == len(roots))
    x = roots[0]
    assert('section1' == x.name)
    assert(x.duration > 0)
    assert(0 == len(x.children))

    ##################################################

    p = ProfilerImpl()
    p.start('section1')
    p.start('section1.1')
    p.start('section1.1.1')   # unterminated
    p.start('section1.1.2')
    p.stop ('section1.1.2')
    p.stop ('section1.1')
    p.stop ('section1')
    p.start('section2')
    p.start('section2.1')
    p.stop ('section2.1')
    p.start('section2.2')
    p.stop ('section2.2')
    p.start('section2.3')
    p.stop ('section2.3')
    p.start('section2.4')
    p.stop ('section2.4')
    p.start('section2.5')
    p.stop ('section2.5')
    p.start('section2.6')
    p.stop ('section2.6')
    p.stop ('section2')

    roots = p.get_root_nodes()
    assert(2 == len(roots))

    s1 = roots[0]
    assert('section1' == s1.name)
    assert(s1.duration > 0)
    assert(1 == len(s1.children) > 0)

    s11 = s1.children[0]
    assert('section1.1' == s11.name)
    assert(1 == len(s11.children))

    s111 = s11.children[0]
    assert('section1.1.1' == s111.name)
    assert(-1 == s111.duration)
    assert(1 == len(s111.children))

    s112 = s111.children[0]
    assert('section1.1.2' == s112.name)
    assert(0 < s112.duration)
    assert(0 == len(s112.children))

    s2 = roots[1]
    assert('section2' == s2.name)
    assert(0 < s2.duration)
    assert(6 == len(s2.children))

    p.dump_data()

if __name__ == "__main__":
    exit(_profiler_test())



#!/usr/bin/env python
# Copyright (c) 2010-2013 Turbulenz Limited
"""
Report metrics on Turbulenz JSON assets.
"""
import logging

from glob import iglob
from optparse import OptionParser, TitledHelpFormatter

from turbulenz_tools.utils.json_stats import analyse_json
from turbulenz_tools.tools.stdtool import simple_options

__version__ = '1.0.0'
__dependencies__ = ['turbulenz_tools.utils.json_stats']

LOG = logging.getLogger(__name__)

def _parser():
    usage = "usage: %prog [options] asset.json [ ... ]"
    description = """\
Report metrics on JSON asset files.

Metrics are:
"keys": number of bytes used by keys.
"punctuation (punctn)": number of bytes used by JSON punctuation, including '[ ] { } " , :'.
"values": number of bytes used by values. For uncompact JSON files this will also include the white space.
"k%": percentage of total size used by the keys.
"p%": percentage of total size used by the punctuation.
"v%": percentage of total size used by the values (and white space).
"# keys": the total number of keys.
"unique": the number of unique keys.
"total": the total asset size in byte.
"gzip": the asset size after gzip compression.
"ratio": the gzip size as a percentage of the uncompressed total size.
"""
    epilog = 'This tool current assumes the JSON asset is compact with no additional white space.'

    parser = OptionParser(description=description, usage=usage, epilog=epilog, formatter=TitledHelpFormatter())
    parser.add_option("--version", action="store_true", dest="output_version", default=False,
                      help="output version number")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", default=False, help="verbose output")
    parser.add_option("-s", "--silent", action="store_true", dest="silent", default=False, help="silent running")
    parser.add_option("-m", "--metrics", action="store_true", dest="metrics", default=False,
                      help="output asset metrics")

    parser.add_option("-H", "--header", action="store_true", dest="header", default=False,
                      help="generate column header")

    return parser

def main():
    (options, args, parser_) = simple_options(_parser, __version__, __dependencies__)

    divider = "+-------------------------+----------------------+---------------+------------------------+"
    if options.header:
        print divider
        print "|    keys: punctn: values |     k%:    p%:    v% | # keys:unique |   total:   gzip: ratio |"
        print divider

    def vadd(a, b):
        return tuple([x + y for (x, y) in zip(a, b)])

    def log((keys, punctuation, values, key_count, unique_count, total_size, compressed_size), f):
        k_percent = keys * 100.0 / total_size
        p_percent = punctuation * 100.0 / total_size
        v_percent = values * 100.0 / total_size
        c_percent = compressed_size * 100.0 / total_size
        print "| %7i:%7i:%7i | %5.1f%%:%5.1f%%:%5.1f%% | %6i:%6i | %7i:%7i:%5.1f%% | %s" % \
            (keys, punctuation, values, k_percent, p_percent, v_percent, key_count, unique_count, \
             total_size, compressed_size, c_percent, f)

    totals = (0, 0, 0, 0, 0, 0, 0)
    for f in args:
        for g in iglob(f):
            stats = analyse_json(g)
            totals = vadd(totals, stats)
            if options.verbose:
                log(stats, g)
    else:
        total_string = 'cumulative total and global ratio'
        if options.verbose:
            print divider

        log(totals, total_string)

        if options.header:
            print divider

    return 0

if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python
# Copyright (c) 2010-2013 Turbulenz Limited

import logging
import os
import tarfile
import simplejson as json

from optparse import OptionParser, TitledHelpFormatter

from turbulenz_tools.tools.stdtool import simple_options

__version__ = '1.0.0'
__dependencies__ = [ ]

LOG = logging.getLogger(__name__)

class DependencyTar(object):
    def __init__(self):
        self.paths = [ ]

    def items(self):
        return self.paths

    def add(self, path, name_):
        self.paths.append(path)

    def close(self):
        pass

def images_in_asset(json_asset):
    """Iterator for all images used in a json asset."""
    images = json_asset.get('images', None)
    if images is not None:
        for _, image in images.iteritems():
            yield image, None

    materials = json_asset.get('materials', None)
    if materials is not None:
        texture_maps = { }
        for _, material in materials.iteritems():
            parameters = material.get('parameters', None)
            if parameters is not None:
                for texture_stage, texture_map in parameters.iteritems():
                    if isinstance(texture_map, (str, unicode)):
                        texture_maps[texture_stage] = True
                        yield texture_map, texture_stage
        LOG.info('contains: %s' % ', '.join(texture_maps.keys()))

def _parser():
    parser = OptionParser(description='Generate a TAR file for binary assets referenced from a JSON asset.',
                          usage='%prog -i input.json -o output.tar [options]',
                          formatter=TitledHelpFormatter())
    parser.add_option("--version", action="store_true", dest="output_version", default=False,
                      help="output version number")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", default=False, help="verbose output")
    parser.add_option("-s", "--silent", action="store_true", dest="silent", default=False, help="silent running")

    parser.add_option("-i", "--input", action="store", dest="input", help="input file to process")
    parser.add_option("-o", "--output", action="store", dest="output", help="output file to process")
    parser.add_option("-a", "--assets", action="store", dest="asset_root", default=".", metavar="PATH",
                      help="path of the asset root")

    parser.add_option("-M", action="store_true", dest="dependency", default=False, help="output dependencies")
    parser.add_option("--MF", action="store", dest="dependency_file", help="dependencies output to file")

    return parser

def main():
    (options, args_, parser_) = simple_options(_parser, __version__, __dependencies__)

    # Cleanly handle asset location
    asset_root = options.asset_root.replace('\\', '/').rstrip('/')

    def _filename(filename):
        if filename[0] == '/':
            return asset_root + filename
        else:
            return asset_root + '/' + filename

    tar_file = DependencyTar()

    LOG.info('%s %s' % (__file__, options.input))
    LOG.info('input: %s' % options.input)

    try:
        with open(options.input, 'r') as source:
            json_asset = json.load(source)
            if not options.dependency:
                tar_file = tarfile.open(options.output, 'w')
            image_map = { }
            (added, missed, skipped) = (0, 0, 0)
            for image_name, texture_stage_ in images_in_asset(json_asset):
                # This is probably a procedural image - skip it
                if image_name not in image_map:
                    # We used to convert .tga -> .png, .cubemap -> .dds, and support dropping mips from dds files.
                    # Currently this is disabled until we integrate the tool with a build process.
                    # Alternatively we expect this conversion to happen in the image pipeline.
                    try:
                        image_path = _filename(image_name)
                        if os.path.exists(image_path):
                            # Actually do the tar add
                            tar_file.add(image_path, image_name)
                            LOG.info('adding: %s' % image_name)
                            image_path = image_path.replace("\\", "/")
                            added += 1
                        else:
                            # We don't mind if files are missing
                            LOG.warning('missing: %s' % image_name)
                            missed += 1
                    except OSError:
                        # We don't mind if files are missing
                        LOG.warning('missing: %s' % image_name)
                        missed += 1
                    image_map[image_name] = 0
                else:
                    LOG.info('skipping: %s' % image_name)
                    skipped += 1
                image_map[image_name] += 1
            tar_file.close()
            LOG.info('output: %s' % options.output)
            LOG.info('report: added %i, missing %i, skipped %i' % (added, missed, skipped))
    except IOError as e:
        LOG.error(e)
        return e.errno
    except Exception as e:
        LOG.critical('Unexpected exception: %s' % e)
        return 1

    if options.dependency:
        if options.dependency_file:
            LOG.info('writing dependencies: %s' % options.dependency_file)
            dep_file = open(options.dependency_file, 'w')

            for dep in tar_file.items():
                dep_file.write("%s\n" % dep)
            dep_file.close()
        else:
            for dep in tar_file.items():
                print dep
            print

    return 0

if __name__ == "__main__":
    exit(main())



# Copyright (c) 2009-2011,2013 Turbulenz Limited

from os.path import getsize as path_getsize
from zlib import compress as zlib_compress
from simplejson import loads as json_loads

__version__ = '1.0.0'

def analyse_json(filename):
    """Utility to return the ratio of key size, punctuation size, and leaf value size."""

    unique_keys = { }

    def __get_size(j):
        """Recurse to generate size."""
        (keys, punctuation, key_count) = (0, 0, 0)
        if isinstance(j, list):
            punctuation += 1 # [
            punctuation += (len(j) - 1) # ,
            for v in j:
                sub_k, sub_p, sub_count = __get_size(v)
                keys += sub_k
                punctuation += sub_p
                key_count += sub_count
            punctuation += 1 # ]
        elif isinstance(j, dict):
            punctuation += 1 # {
            if len(j.keys()) > 1:
                punctuation += (len(j.keys()) - 1) # ,
            for k, v in j.iteritems():
                if k not in unique_keys:
                    unique_keys[k] = True
                key_count += 1
                punctuation += 1 # "
                keys += len(k)
                punctuation += 1 # "
                punctuation += 1 # :
                sub_k, sub_p, sub_count = __get_size(v)
                keys += sub_k
                punctuation += sub_p
                key_count += sub_count
            punctuation += 1 # }
        elif isinstance(j, (str, unicode)):
            punctuation += 1 # "
            punctuation += 1 # "
        return (keys, punctuation, key_count)

    total_size = path_getsize(filename)
    with open(filename, 'r') as f:
        data = f.read()
        j = json_loads(data)

        (keys, punctuation, key_count) = __get_size(j)
        values = total_size - (keys + punctuation)
        unique_count = len(unique_keys.keys())
        compressed_size = len(zlib_compress(data, 6))

        return (keys, punctuation, values, key_count, unique_count, total_size, compressed_size)



# Copyright (c) 2009-2013 Turbulenz Limited
"""
Utilities to manipulate JSON data.
"""

import logging
LOG = logging.getLogger('asset')

def merge_dictionaries(outof, into, prefix='\t'):
    """Merge the dictionary 'outof' into the dictionary 'into'. If matching keys are found and the value is a
    dictionary, then the sub dictionary is merged."""
    for k in outof.keys():
        if k in into:
            if isinstance(outof[k], dict):
                LOG.debug("%sMerging:%s" % (prefix, k))
                into[k] = merge_dictionaries(outof[k], into[k], prefix + '\t')
            else:
                LOG.debug("%sSkipping:%s" % (prefix, k))
        else:
            into[k] = outof[k]
    return into

def float_to_string(f):
    """Unitiliy float encoding which clamps floats close to 0 and 1 and uses %g instead of repr()."""
    if abs(f) < 1e-6:
        return "0"
    elif abs(1 - f) < 1e-6:
        return "1"
    return "%g" % (f)

def metrics(asset):
    """Generate a collection of simple size metrics about the asset."""
    def __approximate_size(num):
        """Convert a file size to human-readable form."""
        for x in [' bytes', 'KB', 'MB', 'GB', 'TB']:
            if num < 1024.0:
                return "%3.1f%s" % (num, x)
            num /= 1024.0

    def __count_nodes(nodes):
        """Recursively count the nodes."""
        num_nodes = len(nodes)
        for n in nodes.itervalues():
            if 'nodes' in n:
                num_nodes += __count_nodes(n['nodes'])
        return num_nodes

    m = { }
    for k in asset.keys():
        if isinstance(asset[k], dict):
            m['num_' + k] = len(asset[k])
        elif isinstance(asset[k], list):
            m['num_' + k] = len(asset[k])
    if 'nodes' in asset:
        m['num_nodes_recurse'] = __count_nodes(asset['nodes'])
    if 'num_geometries' in m and m['num_geometries']:
        m['total_primitives'] = 0
        m['total_positions'] = 0
        m['approximate_size'] = 0
        for _, shape in asset['geometries'].items():
            if 'surfaces' in shape:
                for _, surface in shape['surfaces'].items():
                    m['total_primitives'] += int(surface['numPrimitives'])
            else:
                if 'numPrimitives' in shape:
                    m['total_primitives'] += int(shape['numPrimitives'])
            if 'POSITION' in shape['inputs']:
                position_source = shape['inputs']['POSITION']['source']
                if position_source in shape['sources']:
                    positions = shape['sources'][position_source]
                    m['total_positions'] += len(positions['data']) / positions['stride']
            for _, source in shape['sources'].items():
                m['approximate_size'] += len(source['data']) * 4      # Assume float per vertex element attribute
            if 'triangles' in shape:
                m['approximate_size'] += len(shape['triangles']) * 2  # Assume short per index
            elif 'quads' in shape:
                m['approximate_size'] += len(shape['quads']) * 2      # Assume short per index
        m['average_primitives'] = m['total_primitives'] / m['num_geometries']
        m['average_positions'] = m['total_positions'] / m['num_geometries']
        m['approximate_readable_size'] = __approximate_size(m['approximate_size'])
    return m

def log_metrics(asset):
    """Output the metrics to the log."""
    m = metrics(asset)
    keys = m.keys()
    keys.sort()
    for k in keys:
        LOG.info('%s:%s' % (k, m[k]))



#!/usr/bin/env python
# Copyright (c) 2010-2013 Turbulenz Limited

import logging
import simplejson as json

from optparse import OptionParser, TitledHelpFormatter
from fnmatch import fnmatch

from turbulenz_tools.utils.disassembler import Json2htmlRenderer, Json2txtRenderer, Json2txtColourRenderer, Disassembler
from turbulenz_tools.tools.stdtool import simple_options

__version__ = '1.0.0'
__dependencies__ = ['turbulenz_tools.utils.disassembler']

LOG = logging.getLogger(__name__)

def _parser():
    parser = OptionParser(description='Generate a plain text or html output from a JSON asset. (plain text by default)',
                          usage='%prog -i input.json [-o output.html] [options]',
                          formatter=TitledHelpFormatter())
    parser.add_option("--version", action="store_true", dest="output_version", default=False,
                      help="output version number")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose", default=False, help="verbose output")
    parser.add_option("-s", "--silent", action="store_true", dest="silent", default=False, help="silent running")

    parser.add_option("-i", "--input", action="store", dest="input", help="input file to process")
    parser.add_option("-o", "--output", action="store", dest="output", help="output file to process")

    parser.add_option("-l", "--listcull", action="store", dest="listcull", type="int", default=3, metavar="NUMBER",
                     help="parameter of the list culling size. 0 - show all (defaults to 3)")
    parser.add_option("-c", "--dictcull", action="store", dest="dictcull", type="int", default=3, metavar="NUMBER",
                     help="parameter of the dictionary culling size. 0 - show all (defaults to 3)")
    parser.add_option("-p", "--path", action="store", dest="path", type="str", default=None,
                     help="path of the required node in the asset tree structure (wildcards allowed)")
    parser.add_option("-d", "--depth", action="store", dest="depth", type="int", default=2, metavar="NUMBER",
                     help="parameter of the dictionary and list rendering depth (defaults to 2).")

    parser.add_option("--html", action="store_true", dest="html", default=False,
                      help="output in html format")
    parser.add_option("--txt", action="store_true", dest="txt", default=False,
                      help="output in plain text format")
    parser.add_option("--color", action="store_true", dest="color", default=False,
                      help="option to turn on the coloured text output")

    return parser

def main():
    (options, args_, parser_) = simple_options(_parser, __version__, __dependencies__)

    source_file = options.input

    LOG.info('%s %s' % (__file__, source_file))
    LOG.info('input: %s' % source_file)

    try:
        with open(source_file, 'r') as source:
            json_asset = json.load(source)

            def find_node(nodes, sub_asset):
                for (k, v) in sub_asset.iteritems():
                    if fnmatch(k, nodes[0]):
                        if len(nodes) == 1:
                            yield (k, sub_asset[k])
                        elif isinstance(v, dict):
                            for n in find_node(nodes[1:], sub_asset[k]):
                                yield n
                            for n in find_node(nodes, sub_asset[k]):
                                yield n

            if options.path:
                node_list = options.path.split('/')

            if options.html:
                renderer = Json2htmlRenderer()
            elif options.color:
                renderer = Json2txtColourRenderer()
            else:
                renderer = Json2txtRenderer()

            disassembler = Disassembler(renderer, options.listcull, options.dictcull, options.depth)
            expand = True

            if options.output:
                with open(options.output, 'w') as target:
                    if options.path:
                        for name, node in find_node(node_list, json_asset):
                            target.write(name + ': ' + disassembler.mark_up_asset({name: node}, expand))
                            target.write('\n')
                    else:
                        target.write(disassembler.mark_up_asset({'root': json_asset}, expand))
                        target.write('\n')
            elif options.color:
                if options.path:
                    for name, node in find_node(node_list, json_asset):
                        print '\033[;31m' + name + '\033[;m' + ': ' + disassembler.mark_up_asset({name: node}, expand),
                else:
                    print disassembler.mark_up_asset({'root': json_asset}, expand)
            else:
                if options.path:
                    for name, node in find_node(node_list, json_asset):
                        print name + ': ' + disassembler.mark_up_asset({name: node}, expand),
                else:
                    print disassembler.mark_up_asset({'root': json_asset}, expand)

    except IOError as e:
        LOG.error(e)
        return e.errno
    except Exception as e:
        LOG.critical('Unexpected exception: %s' % e)
        return 1
    return 0

if __name__ == "__main__":
    exit(main())



# Copyright (c) 2012-2013 Turbulenz Limited

"""
This file contains all of the code generation, formatting and default
templates for the build tools.  This includes the set of variables
used to render the html templates, the format of dependency
information and the set of shared options across the code build tools.
"""

from turbulenz_tools.utils.dependencies import find_file_in_dirs
from turbulenz_tools.utils.profiler import Profiler
from turbulenz_tools.tools.toolsexception import ToolsException
from turbulenz_tools.tools.templates import read_file_utf8

import os.path
import glob
from re import compile as re_compile
from logging import getLogger

__version__ = '1.1.4'

LOG = getLogger(__name__)

############################################################

DEFAULT_HTML_TEMPLATE = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">
<html>
<head>
    <title>
        /*{% block tz_app_title %}*//*{{ tz_app_title_var }}*//*{% endblock %}*/
    </title>
    <meta http-equiv="Content-Type" content="text/html;charset=utf-8" >
    <style type="text/css">
html, body, div, span, object, iframe, h1, h2, p, a, img, ul, li, fieldset, form, label, legend, table, thead, tbody, tfoot, tr, th, td {
    border: 0;
    font-size: 100%;
    margin: 0;
    outline: 0;
    padding: 0;
    vertical-align: baseline;
}
    </style>
    <!-- block tz_app_header -->
    /*{% block tz_app_header %}*//*{% endblock %}*/
    <!-- end tz_app_header -->
</head>
<body style="background:#B4B4B4;font:normal normal normal 13px/1.231 Helvetica,Arial,sans-serif;text-shadow:1px 1px #F9F8F8;">
    <div id="titlebar" style="position:fixed;height:65px;top:0;right:0;left:0;">
        <strong style="font-size:24px;line-height:64px;margin:16px;">
            <!-- block tz_app_title_name -->
            /*{% block tz_app_title_name %}*/
            /*{{ tz_app_title_name_var }}*/
            /*{% endblock %}*/
            <!-- end tz_app_title_name -->
        </strong>
        <div id="titlelogo"
             style="float:right;width:27px;height:27px;margin:18px 24px;">
        </div>
    </div>
    <div id="sidebar"
         style="background:#B4B4B4;position:fixed;width:303px;top:65px;left:0;">
        <!-- block tz_app_html_controls -->
        /*{% block tz_app_html_controls %}*/
        /*{% endblock %}*/
        <!-- end tz_app_html_controls -->
    </div>
    <div id="engine" style="background:#939393;position:fixed;top:65px;
                            bottom:0;right:0;left:303px;
                            border-left:1px solid #898989;">
        <!--
          HTML to create a plugin or canvas instance.
          Supplied by 'tz_engine_div' variable.
        -->
        /*{{ tz_engine_div }}*/
    </div>

    <!-- begin 'tz_include_js' variable -->
    /*{{ tz_include_js }}*/
    <!-- end 'tz_include_js' variable -->

    <script type="text/javascript">
      // ----------------------------------------
      // Embedded code and startup code.
      // Supplied by 'tz_startup_code' variable.
      // ----------------------------------------
      /*{{ tz_startup_code }}*/
    </script>

</body>
</html>
"""

############################################################

def default_parser_options(parser):
    """
    Command line options shared by make*.py tools
    """

    parser.add_option("--version", action="store_true", dest="output_version",
                      default=False, help="output version number")
    parser.add_option("-v", "--verbose", action="store_true", dest="verbose",
                      default=False, help="verbose output")
    parser.add_option("-s", "--silent", action="store_true", dest="silent",
                      default=False, help="silent running")

    # Input / Output (input .html and .js files don't need a prefix)
    parser.add_option("-o", "--output", action="store", dest="output",
                      help="output file to process")
    parser.add_option("-t", "--templatedir", action="append", dest="templatedirs",
                      default=[], help="template directory (multiple allowed)")

    # Dependency generation
    parser.add_option("-M", "--dependency", action="store_true",
                      dest="dependency", default=False,
                      help="output dependencies")
    parser.add_option("--MF", action="store", dest="dependency_file",
                      help="dependencies output to file")

    # "use strict" options
    parser.add_option("--use-strict", action="store_true", dest="use_strict",
                      default=False, help='enforce "use strict"; statement. '
                      'This adds a single "use strict"; line at the top of the '
                      'JavaScript code.')
    parser.add_option("--include-use-strict", action="store_true",
                      dest="include_use_strict", default=False,
                      help='don\'t strip out "use strict"; statements. '
                      'By default all "use strict"; statements are removed '
                      'from the output file.')

    # Hybrid
    parser.add_option("--hybrid", action="store_true", dest="hybrid",
                      default=False, help="canvas, canvas_dev modes only. "
                      "Start up a plugin as well as a canvas-based "
                      "TurbulenzEngine. The plugin will be available as "
                      "TurbulenzEnginePlugin.")

    # Profiling
    def _enable_profiler(_options, _opt_str, _value, _parser):
        Profiler.enable()
    parser.add_option("--profile", action="callback", callback=_enable_profiler,
                      help="enable the collection and output of profiling "
                      "information")

    # Injecting files
    parser.add_option("--no-inject", action="store_true", dest="noinject",
                      default=False, help="Don't inject default library files")

############################################################

def render_js(context, options, templates_js, inject_js):
    """
    Renders the templates in templates_js, as if the first template
    began with include declarations for each of the files in
    inject_js.  Returns the result of rendering, and the list of
    includes that were not inlined.  (rendered_js, inc_js)

    For dev modes, the list of includes is returned in inc_js as
    relative paths from the output file.  For release modes, includes
    are all inlined (inc_js == []).
    """

    regex_use_strict = re_compile('"use strict";')

    out = []
    inc_js = []
    outfile_dir = os.path.abspath(os.path.dirname(options.output)) + os.sep

    includes_seen = []

    # Any headers

    if options.use_strict:
        out.append('"use strict";')

    if options.mode in [ 'plugin', 'canvas' ]:
        out.append('(function () {')

    # Functions for handling includes

    def _find_include_or_error(name):
        try:
            f = find_file_in_dirs(name, options.templatedirs)
        except Exception, ex:
            raise ToolsException(str(ex))
        if f is None:
            raise ToolsException("No file '%s' in any template dir" % name)
        LOG.info(" resolved '%s' to path '%s'" % (name, f))
        return f

    def handle_javascript_dev(name):
        file_path = _find_include_or_error(name)
        if file_path in includes_seen:
            LOG.info(" include '%s' (%s) already listed" % (name, file_path))
            return ""
        includes_seen.append(file_path)

        # Calculate relative path name
        # rel_path = file_path.replace(outfile_dir, '').replace('\\', '/')
        # if rel_path == file_path:
        #     raise ToolsException("Included file '%s' found at '%s', which is "
        #                          "not in a child directory of the output file "
        #                          "'%s' in directory %s" % (name, file_path,
        #                                                    options.output,
        #                                                    outfile_dir))
        rel_path = os.path.relpath(file_path, outfile_dir).replace('\\', '/')

        inc_js.append(rel_path)
        return ""

    def handle_javascript_release(name):
        if options.stripdebug and os.path.basename(name) == "debug.js":
            LOG.warning("App attempting to include debug.js.  Removing.")
            return ""
        file_path = _find_include_or_error(name)
        if file_path in includes_seen:
            LOG.info(" include '%s' (%s) already listed" % (name, file_path))
            return ""
        includes_seen.append(file_path)

        d = read_file_utf8(file_path)

        if options.include_use_strict:
            return d
        else:
            # strip out any "use strict"; lines
            return regex_use_strict.sub('', d)

    if options.mode in [ 'plugin', 'canvas' ]:
        handle_javascript = handle_javascript_release
    else:
        handle_javascript = handle_javascript_dev
    context['javascript'] = handle_javascript

    # Inject any includes at the start, either embedding them or
    # adding to the inc_js list.

    for inj in inject_js:
        js_line = handle_javascript(inj)
        if js_line:
            out.append(js_line)

    # Render templates

    out += [t.render(context) for t in templates_js]
    del context['javascript']

    # Any footer code

    if options.mode == 'plugin':
        out.append("""
    if (!TurbulenzEngine.onload)
    {
        window.alert("Entry point 'TurbulenzEngine.onload' must be defined.");
        return;
    }
    TurbulenzEngine.onload.call(this);
}());""")

    if options.mode == 'canvas':
        out.append('window.TurbulenzEngine = TurbulenzEngine;}());')

    # Combine all parts into a single string

    return ("\n".join(out), inc_js)

def render_js_extract_includes(context, options, templates_js, injects):
    """
    Renders the templates in templates_js against the given context
    and just collects the set of 'javascript('...')' includes.  Will
    optionally handle a list of files to be injected.

    Returns an array of absolute paths, removing duplicates.
    """

    includes = []

    def _find_in_dirs_or_error(name):
        file_path = find_file_in_dirs(name, options.templatedirs)
        if file_path is None:
            raise ToolsException("No file '%s' in any template dir" % name)
        if file_path in includes:
            LOG.info(" include '%s' (%s) already listed" % (name, file_path))
            return
        LOG.info(" resolved '%s' to path '%s'" % (name, file_path))
        includes.append(file_path)

    # In release mode, filter out debug.js

    if options.mode in [ 'plugin', 'canvas' ] and options.stripdebug:
        _do_find_in_dirs_or_error = _find_in_dirs_or_error
        # pylint: disable=E0102
        def _find_in_dirs_or_error(name):
            if os.path.basename(name) == "debug.js":
                LOG.warning("App attempting to include debug.js.  Removing.")
                return
            _do_find_in_dirs_or_error(name)
        # pylint: enable=E0102

    # Deal with any injects

    for i in injects:
        _find_in_dirs_or_error(i)

    # Use the templating engine to deal with remaining includes

    def handle_javascipt_extract_includes(name):
        _find_in_dirs_or_error(name)
        return ""

    context['javascript'] = handle_javascipt_extract_includes

    for t in templates_js:
        t.render(context)

    del context['javascript']

    return includes

############################################################

def output_dependency_info(dependency_file, output_file, dependencies):
    """
    This dependency write outputs dependency information in a format
    consistent with the GCC -M flags.
    """

    try:
        with open(dependency_file, "wb") as f:
            f.write(output_file)
            f.write(" ")
            f.write(dependency_file)
            f.write(" : \\\n")
            for d in dependencies:
                f.write("    ")
                f.write(d)
                f.write(" \\\n")
            f.write("\n\n")
            for d in dependencies:
                f.write(d)
                f.write(" :\n\n")
    except IOError:
        raise ToolsException("failed to write file: %s" % dependency_file)

############################################################

def context_from_options(options, title):

    # Sanity check

    if options.hybrid:
        if options.mode not in [ 'canvas', 'canvas-debug' ]:
            raise ToolsException("--hybrid option available only in canvas and "
                                 "canvas_dev modes")

    # Set up the context

    context = {}
    context['tz_app_title_name_var'] = title
    context['tz_app_title_var'] = title

    context['tz_development'] = options.mode in [ 'plugin-debug', 'canvas-debug' ]
    context['tz_canvas'] = options.mode in [ 'canvas', 'canvas-debug' ]
    context['tz_hybrid'] = options.hybrid

    return context

############################################################

def inject_js_from_options(options):
    """
    Given the build options, find (if necessary), all includes that
    must be injected for canvas mode to work.  This is done by
    searching for webgl_engine_file in any of the
    template directories, and collecting the list of all .js files
    that reside there.
    """

    inject_list = []

    if options.noinject:
        return inject_list

    mode = options.mode

    # Put debug.js at the top (if in debug mode), and ALWAYS include
    # vmath.js

    if mode in [ 'plugin-debug', 'canvas-debug' ] or not options.stripdebug:
        inject_list.append('jslib/debug.js')
    inject_list.append('jslib/vmath.js')

    # Include webgl includes in canvas mode

    if mode in [ 'canvas', 'canvas-debug' ]:
        LOG.info("Looking for jslib/webgl ...")

        webgl_engine_file = 'jslib/webgl/turbulenzengine.js'
        webgl_engine_dir = os.path.dirname(webgl_engine_file)

        # Find absolute path of webgl_engine_file

        webgl_abs_path = None

        for t in options.templatedirs:
            p = os.path.join(t, webgl_engine_file)
            if os.path.exists(p):
                webgl_abs_path = os.path.dirname(p)
                LOG.info("Found at: %s" % webgl_abs_path)
                break

        if webgl_abs_path is None:
            raise ToolsException("No '%s' in any template dir" \
                                     % webgl_engine_file)

        webgl_abs_files = glob.glob(webgl_abs_path + "/*.js")
        inject_list += [ 'jslib/utilities.js',
                         'jslib/aabbtree.js',
                         'jslib/observer.js' ]
        inject_list += \
            [ webgl_engine_dir + "/" + os.path.basename(f) for f in webgl_abs_files ]

    return inject_list

############################################################

def default_add_code(options, context, rendered_js, inc_js):
    """
    Add:
      tz_engine_div
      tz_include_js
      tz_startup_code
    to the context, based on the values of options
    """

    # pylint: disable=W0404
    from turbulenz_tools.version import SDK_VERSION as engine_version
    # pylint: enable=W0404
    engine_version_2 = ".".join(engine_version.split(".")[0:2])

    outfile_dir = os.path.dirname(options.output)
    if options.mode in [ 'plugin', 'canvas' ]:
        codefile_rel = os.path.relpath(options.codefile, outfile_dir).replace('\\', '/')

    #
    # tz_engine_div and tz_engine_2d_div
    #

    tz_engine_div = ""
    if options.mode in [ 'canvas', 'canvas-debug' ]:

        tz_engine_div += """
        <canvas id="turbulenz_game_engine_canvas" moz-opaque="true" tabindex="1">
            Sorry, but your browser does not support WebGL or does not have it
            enabled.  To get a WebGL-enabled browser, please see:<br/>
            <a href="http://www.khronos.org/webgl/wiki/Getting_a_WebGL_Implementation" target="_blank">
                Getting a WebGL Implementation
            </a>
        </canvas>

        <script type="text/javascript">
            var canvasSupported = true;
            (function()
            {
                var contextNames = ["webgl", "experimental-webgl"];
                var context = null;
                var canvas = document.createElement('canvas');

                document.body.appendChild(canvas);

                for (var i = 0; i < contextNames.length; i += 1)
                {
                    try {
                        context = canvas.getContext(contextNames[i]);
                    } catch (e) {}

                    if (context) {
                        break;
                    }
                }
                if (!context)
                {
                    canvasSupported = false;
                    window.alert("Sorry, but your browser does not support WebGL or does not have it enabled.");
                }

                document.body.removeChild(canvas);
            }());
            var TurbulenzEngine = {};
        </script>"""

        tz_engine_2d_div = """
        <canvas id="turbulenz_game_engine_canvas" moz-opaque="true" tabindex="1">
        </canvas>

        <script type="text/javascript">
            var canvasSupported = true;
            (function()
            {
                var canvas = document.createElement("canvas");
                document.body.appendChild(canvas);
                if (!canvas.getContext("2d"))
                {
                    canvasSupported = false;
                    window.alert("Sorry, but your browser does not support 2D Canvas or does not have it enabled.");
                }
                document.body.removeChild(canvas);
            }());
            var TurbulenzEngine = {};
        </script>"""


    if options.mode in [ 'plugin', 'plugin-debug' ] or options.hybrid:

        tz_engine_div += """
        <script type="text/javascript">
            if (window.ActiveXObject)
            {
                document.write('<object id="turbulenz_game_loader_object" classid="CLSID:49AE29B1-3E7D-4f62-B3D2-D6F7C7BEE728" width="100%" height="100%">');
                document.write('<param name="type" value="application/vnd.turbulenz" \/>');
                document.write('<p>You need the Turbulenz Engine for this.');
                document.write('<\/p>');
                document.write('<\/object>');
            }
            else
            {
                // browser supports Netscape Plugin API
                document.write('<object id="turbulenz_game_loader_object" type="application/vnd.turbulenz" width="100%" height="100%">');
                document.write('<p>You need the Turbulenz Engine for this.');
                document.write('<\/p>');
                document.write('<\/object>');
            }"""

        if options.mode == 'plugin-debug':
            tz_engine_div += """
            // If IE
            if (navigator.appName === "Microsoft Internet Explorer")
            {
                window.alert("Sorry, but this sample does not run in development mode in Internet Explorer.");
            }
            var TurbulenzEngine = {
                VMath: null
            };"""

        tz_engine_div += """
        </script>"""

        tz_engine_2d_div = tz_engine_div

    #
    # tz_include_js
    #

    if options.mode in [ 'plugin-debug', 'canvas-debug' ]:
        inc_lines = [ '<script type="text/javascript" src="%s"></script>' % js \
                          for js in inc_js ]
        tz_include_js = "\n".join(inc_lines)
    elif options.mode == 'canvas':
        tz_include_js = '\n<script type="text/javascript" src="%s"></script>' \
            % codefile_rel
    else:
        tz_include_js = "\n"

    #
    # tz_startup_code
    #

    tz_startup_find_best_version_fn = """

            function findBestVersion(request, availableVersions)
            {
                var reqNumbers = request.split(".");
                var candidate;

                for (var vIdx = 0; vIdx < availableVersions.length; vIdx += 1)
                {
                    var ver = availableVersions[vIdx];
                    var verNumbers = ver.split(".");

                    // Check the version has the correct major and minor

                    if ((verNumbers[0] !== reqNumbers[0]) ||
                        (verNumbers[1] !== reqNumbers[1]))
                    {
                        continue;
                    }

                    // If there is already a candidate, compare point and build

                    if (candidate)
                    {
                        if (verNumbers[2] > candidate[2])
                        {
                            candidate = verNumbers;
                            continue;
                        }
                        if ((verNumbers[2] === candidate[2]) &&
                            (verNumbers[3] > candidate[3]))
                        {
                            candidate = verNumbers;
                            continue;
                        }
                    }
                    else
                    {
                        candidate = verNumbers;
                    }
                }

                if (candidate)
                {
                    candidate = candidate.join(".");
                }
                return candidate;
            }"""

    tz_startup_plugin_unload_code = """

            // Engine unload
            var previousOnBeforeUnload = window.onbeforeunload;
            window.onbeforeunload = function ()
            {
                try {
                    loader.unloadEngine();
                } catch (e) {
                }
                if (previousOnBeforeUnload) {
                    previousOnBeforeUnload.call(this);
                }
            };"""

    tz_startup_plugin_check_and_load_code = tz_startup_find_best_version_fn + """

            var now = Date.now || function () { return new Date().valueOf(); };
            var loadDeadline = now() + 5 * 1000;  // 5 seconds
            var loadInterval = 500;               // 0.5 seconds

            var attemptLoad = function attemptLoadFn()
            {
                // Check plugin and load engine
                var err = 0;
                if (!loader) {
                    err = "no loader DOM element";
                }
                if (err === 0 &&
                    !loader.loadEngine &&
                    loader.hasOwnProperty &&
                    !loader.hasOwnProperty('loadEngine')) {
                    err = "loader has no 'loadEngine' property";
                }
                if (err === 0 &&
                    !loader.getAvailableEngines &&
                    !loader.hasOwnProperty('getAvailableEngines')) {
                    err = "no 'getAvailableEngines'. Plugin may be "
                            + "an older version.";
                }

                if (err === 0)
                {
                    var availableEngines = loader.getAvailableEngines();
                    var samplesVersion = '""" + engine_version_2 + """';
                    var requestVersion =
                        findBestVersion(samplesVersion, availableEngines);
                    if (!requestVersion)
                    {
                        err = "No engines installed that are compatible with "
                                + "version " + samplesVersion;
                    }
                    else
                    {
                        config.version = requestVersion;
                    }
                }

                if (err === 0)
                {
                    // Plugin is in place
                    if (!loader.loadEngine(config))
                    {
                        window.alert("Call to loadEngine failed");
                    }
                    return;
                }

                // Continue to wait for the plugin
                if (loadDeadline >= now()) {
                    window.setTimeout(attemptLoad, loadInterval);
                } else {
                    window.alert("No Turbulenz Loader found ("+err+")");
                }
            };
            attemptLoad();"""

    if options.mode == 'plugin-debug':
        tz_startup_code = "\n" + rendered_js.lstrip('\n') + """

        // Engine startup
        window.onload = function ()
        {
            var loader =
                document.getElementById('turbulenz_game_loader_object');
            var appEntry = TurbulenzEngine.onload;
            var appShutdown = TurbulenzEngine.onunload;
            var appMathDevice = TurbulenzEngine.VMath;
            if (!appEntry)
            {
                window.alert("TurbulenzEngine.onload has not been set");
                return;
            }
            var progressCB = function progressCBFn(msg)
            {
                if ('number' !== typeof msg) {
                    window.alert("Error during engine load: " + msg);
                    return;
                }
                // Progress update here
            };
            var config = {
                run: function runFn(engine) {
                    TurbulenzEngine = engine;
                    TurbulenzEngine.onload = appEntry;
                    TurbulenzEngine.onunload = appShutdown;
                    TurbulenzEngine.VMath = appMathDevice;
                    engine.setTimeout(appEntry, 0);
                },
                progress: progressCB
            };""" + tz_startup_plugin_unload_code + """

            """ + tz_startup_plugin_check_and_load_code + """

        };  // window.onload()"""

    elif options.mode == 'plugin':
        tz_startup_code = """

        window.onload = function ()
        {
            var loader = document.getElementById('turbulenz_game_loader_object');
            var config = {
                url: '""" + codefile_rel + """'
            };""" + tz_startup_plugin_unload_code + """

            """ + tz_startup_plugin_check_and_load_code + """

        };  // window.onload()"""

    else:
        tz_startup_code = ""
        if options.hybrid:
            tz_startup_code += """
        var TurbulenzEnginePlugin = null;"""
        if options.mode == 'canvas-debug':
            tz_startup_code += "\n" + rendered_js.lstrip('\n') + "\n"

        tz_startup_code += """
        // Engine startup
        window.onload = function ()
        {
            var appEntry = TurbulenzEngine.onload;
            var appShutdown = TurbulenzEngine.onunload;
            if (!appEntry) {
                window.alert("TurbulenzEngine.onload has not been set");
                return;
            }

            var canvas =
                document.getElementById('turbulenz_game_engine_canvas');"""

        if options.hybrid:
            tz_startup_code += """
            var loader =
                document.getElementById('turbulenz_game_loader_object');"""

        tz_startup_code += """

            var startCanvas = function startCanvasFn()
            {
                if (canvas.getContext && canvasSupported)
                {
                    TurbulenzEngine = WebGLTurbulenzEngine.create({
                        canvas: canvas,
                        fillParent: true
                    });

                    if (!TurbulenzEngine) {
                        window.alert("Failed to init TurbulenzEngine (canvas)");
                        return;
                    }

                    TurbulenzEngine.onload = appEntry;
                    TurbulenzEngine.onunload = appShutdown;
                    appEntry()
                }
            }"""

        # Unloading code

        tz_startup_code += """

            var previousOnBeforeUnload = window.onbeforeunload;
            window.onbeforeunload = function ()
            {
                if (TurbulenzEngine.onunload) {
                    TurbulenzEngine.onunload.call(this);
                }"""
        if options.hybrid:
            tz_startup_code += """
                if (loader.unloadEngine) {
                    loader.unloadEngine();
                }
                if (previousOnBeforeUnload) {
                    previousOnBeforeUnload.call(this);
                }"""
        tz_startup_code += """
            };  // window.beforeunload"""


        # In hybrid mode, nothing can start until the engine is
        # loaded so wrap the canvas startup code in loader calls,
        # otherwise just call it immediately.

        if options.hybrid:
            tz_startup_code += """
            var config = {
                run: function (engine)
                {
                    TurbulenzEnginePlugin = engine;
                    startCanvas();
                },
            };

            if (!loader || !loader.loadEngine || !loader.loadEngine(config))
            {
                window.alert("Failed to load Turbulenz Engine.");
                return;
            }"""

        else:
            tz_startup_code += """

            startCanvas();"""

        tz_startup_code += """
        };  // window.onload()
"""

    context['tz_engine_div'] = tz_engine_div
    context['tz_engine_2d_div'] = tz_engine_2d_div
    context['tz_include_js'] = tz_include_js
    context['tz_startup_code'] = tz_startup_code
    context['tz_sdk_version'] = engine_version
